;; guile-lib
;; Copyright (C) 2004 Andy Wingo <wingo at pobox dot com>
;; Copyright (C) 2001,2002 Oleg Kiselyov <oleg at pobox dot com>

;; This file is based on SSAX's SSAX.scm, and is in the public domain.

;;; Commentary:
;;
;; Doc me!
;;
;;; Code:

(define-module (xml texinfo)
  #:use-module (xml transform)
  #:use-module (text lex-simple)
  #:use-module (io string)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:export (texinfo->sxml))


(define (parser-error port message . rest)
  (apply error port message rest))

;========================================================================
;				Data Types

; TAG-KIND
;	A symbol 'EMPTY, 'START, 'LINE, 'END

; TAG-NAME
;	A a single symbol.

; ELEM-CONTENT-MODEL
;	A symbol:
;	ENVIRON	  - The tag is an environment tag, expect `@end foo'.
;	EMPTY-COMMAND - no content, and no @end is coming
;	EOL-ARGS  - Attributes := CDATA to end of line
;	EOL-TEXT  - Attributes end at end of line, but are parsed
;       INLINE-ARGS - Attribute is inline
;       INLINE-TEXT - Attribute is inline

; ATTLIST
;	An ordered collection of (NAME . VALUE) pairs, where NAME is
;	a TAG-NAME. The collection is an ADT

; STR-HANDLER
;	A procedure of three arguments: STRING1 STRING2 SEED
;	returning a new SEED
;	The procedure is supposed to handle a chunk of character data
;	STRING1 followed by a chunk of character data STRING2.
;	STRING2 is a short string, often "\n" and even ""

; INCLUDES
;	A stack of strings indicating which files are being included.
;       An attempt to include a file in this list will result in an
;       error: violation of nonrecursion.

; TOKEN -- a record

; This record represents a markup, which is either a stop or an end tag.
;
;	kind -- a TAG-KIND
;	head -- a TAG-NAME. Only 'START For xml-tokens of kinds 'COMMENT and
;		'CDSECT, the head is #f
;
; For example,
;	<P>     => kind='START, head='P
;	</P>    => kind='END, head='P
;	@dots{} => kind='EMPTY-EL, head='BR
; 
; Character references are not represented by xml-tokens as these references
; are transparently resolved into the corresponding characters.

(define (make-token kind head) (cons kind head))
(define token? pair?)
(define token-kind car)
(define token-head cdr)

;; The @ is for attributes
(define (space-significant? command)
  (memq command
        '(example smallexample verbatim lisp smalllisp @)))

;; Like a DTD for texinfo
(define (command-spec command)
  (or (assq command texinfo-command-specs)
      (parser-error #f "Unknown command" command)))

(define (inline-content? content)
  (or (eq? content 'INLINE-TEXT) (eq? content 'INLINE-ARGS)))

(define texinfo-command-specs
  ;; name elem-content-model attrs
  ;; attrs has the same format as the formals for a lambda
  ;; #f for attrs means no args (expect EOL or {}),
  ;; #t means put args as leaves of element
  '(;; Inline text elements
    (bold               INLINE-TEXT . #t)
    (copyright          INLINE-TEXT . #f)
    (sample             INLINE-TEXT . #t)
    (samp               INLINE-TEXT . #t)
    (code               INLINE-TEXT . #t)
    (kbd                INLINE-TEXT . #t)
    (key                INLINE-TEXT . #t)
    (var                INLINE-TEXT . #t)
    (env                INLINE-TEXT . #t)
    (file               INLINE-TEXT . #t)
    (command            INLINE-TEXT . #t)
    (option             INLINE-TEXT . #t)
    (dfn                INLINE-TEXT . #t)
    (cite               INLINE-TEXT . #t)
    (acro               INLINE-TEXT . #t)
    (url                INLINE-TEXT . #t)
    (email              INLINE-TEXT . #t)
    (emph               INLINE-TEXT . #t)
    (strong             INLINE-TEXT . #t)
    (sample             INLINE-TEXT . #t)
    (sc                 INLINE-TEXT . #t)
    (titlefont          INLINE-TEXT . #t)
    (*braces*           INLINE-TEXT . #t)
    (asis               INLINE-TEXT . #t)

    ;; Inline args elements
    (value              INLINE-ARGS . (key))
    (ref                INLINE-ARGS . (node . rest))
    (xref               INLINE-ARGS . (node . rest))
    (pxref              INLINE-ARGS . (node . rest))
    (uref               INLINE-ARGS . (url . rest))
    (anchor             INLINE-ARGS . (name))
    (dots               INLINE-ARGS . #f)
    (bullet             INLINE-ARGS . #f)

    ;; EOL args elements
    (node               EOL-ARGS . (this . rest))
    (c                  EOL-ARGS . #t)
    (comment            EOL-ARGS . #t)
    (setchapternewpage  EOL-ARGS . #t)
    (sp                 EOL-ARGS . #t)
    (page               EOL-ARGS . #f)
    (vskip              EOL-ARGS . #t)

    (para               PARAGRAPH . #f)
    (item               ITEM . heading)
    (itemx              EOL-TEXT . #t)

    (set                EOL-TEXT . #t)
    (center             EOL-TEXT . #t)
    (title              EOL-TEXT . #t)
    (subtitle           EOL-TEXT . #t)
    (author             EOL-TEXT . #t)
    (chapter            EOL-TEXT . #t)
    (section            EOL-TEXT . #t)
    (appendix           EOL-TEXT . #t)
    (appendixsec        EOL-TEXT . #t)
    (unnumbered         EOL-TEXT . #t)
    (unnumberedsec      EOL-TEXT . #t)
    (subsection         EOL-TEXT . #t)
    (subsubsection      EOL-TEXT . #t)
    (appendixsubsec     EOL-TEXT . #t)
    (appendixsubsubsec  EOL-TEXT . #t)
    (unnumberedsubsec   EOL-TEXT . #t)
    (unnumberedsubsubsec EOL-TEXT . #t)

    (noindent           EMPTY-COMMAND . #t)

    (*TOP*              ENVIRON . (title))
    (*ENVIRON-ATTRS*    EOL-TEXT . #t)

    (ifinfo             ENVIRON . #f)
    (iftex              ENVIRON . #f)
    (ifhtml             ENVIRON . #f)
    (ifplaintext        ENVIRON . #f)
    (titlepage          ENVIRON . #f)
    (menu               ENVIRON . #f)
    
    (example            ENVIRON . #f)
    (smallexample       ENVIRON . #f)
    (display            ENVIRON . #f)
    (smalldisplay       ENVIRON . #f)
    (verbatim           ENVIRON . #f)
    (format             ENVIRON . #f)
    (smallformat        ENVIRON . #f)
    (lisp               ENVIRON . #f)
    (smalllisp          ENVIRON . #f)
    (cartouche          ENVIRON . #f)
    (quotation          ENVIRON . #f)

    (table              TABLE-ENVIRON . (how))
    (itemize            TABLE-ENVIRON . (how))
    (enumerate          TABLE-ENVIRON . (how))
    (ftable             TABLE-ENVIRON . (how))
    (vtable             TABLE-ENVIRON . (how))

    (deftp              ENVIRON . #t)
    (defcv              ENVIRON . #t)
    (defivar            ENVIRON . #t)
    (deftypeivar        ENVIRON . #t)
    (defop              ENVIRON . #t)
    (deftypeop          ENVIRON . #t)
    (defmethod          ENVIRON . #t)
    (deftypemethod      ENVIRON . #t)
    (defopt             ENVIRON . #t)
    (defvr              ENVIRON . #t)
    (defvar             ENVIRON . #t)
    (deftypevr          ENVIRON . #t)
    (deftypevar         ENVIRON . #t)
    (deffn              ENVIRON . #t)
    (deffnx             EOL-TEXT . #t)
    (deftypefn          ENVIRON . #t)
    (defspec            ENVIRON . #t)
    (defmac             ENVIRON . #t)
    (defun              ENVIRON . #t)
    (deftypefun         ENVIRON . #t)))

;-------------------------
; Utilities

; Test if a string is made of only whitespace
; An empty string is considered made of whitespace as well
(define (string-whitespace? str)
  (let ((len (string-length str)))
    (cond
     ((zero? len) #t)
     ((= 1 len) (char-whitespace? (string-ref str 0)))
     ((= 2 len) (and (char-whitespace? (string-ref str 0))
		     (char-whitespace? (string-ref str 1))))
     (else
      (let loop ((i 0))
	(or (>= i len)
	    (and (char-whitespace? (string-ref str i))
		 (loop (+ 1 i)))))))))

;; allows eof
(define read-eof-breaks '(*eof* #\return #\newline))
(define (read-eof-line port)
  (if (eof-object? (peek-char port))
      (peek-char port)
      (let* ((line (next-token '() read-eof-breaks
                               "reading a line" port))
             (c (read-char port)))	; must be either \n or \r or EOF
        (if (and (eq? c #\return) (eq? (peek-char port) #\newline))
            (read-char port))		; skip \n that follows \r
        line)))

; Like let* but allowing for multiple-value bindings
(define-macro (let*-values bindings . body)
  (if (null? bindings) (cons 'begin body)
      (apply
       (lambda (vars initializer)
	 (let ((cont 
		(cons 'let*-values
		      (cons (cdr bindings) body))))
	   (cond
	    ((not (pair? vars))		; regular let case, a single var
	     `(let ((,vars ,initializer)) ,cont))
	    ((null? (cdr vars))		; single var, see the prev case
	     `(let ((,(car vars) ,initializer)) ,cont))
	   (else			; the most generic case
	    `(call-with-values (lambda () ,initializer)
	      (lambda ,vars ,cont))))))
       (car bindings))))

(define ascii->char integer->char)

;========================================================================
;		Lower-level parsers and scanners
;
; They deal with primitive lexical units (Names, whitespaces, tags)
; and with pieces of more generic productions. Most of these parsers
; must be called in appropriate context. For example, complete-start-tag
; must be called only when the start-tag has been detected and its GI
; has been read.

(define (skip-whitespace port)
  (skip-while '(#\space #\tab #\return #\newline) port))

(define (skip-horizontal-whitespace port)
  (skip-while '(#\space #\tab) port))

; command ::= Letter+
; Read a command starting from the current position in the PORT and
; return it as a symbol.
(define (read-command port)
  (let ((first-char (peek-char port)))
    (or (char-alphabetic? first-char)
        (parser-error port "Nonalphabetic @-command char: '" first-char "'")))
  (string->symbol
    (next-token-of
      (lambda (c)
        (cond
          ((eof-object? c) #f)
          ((char-alphabetic? c) c)
          (else #f)))
      port)))

; procedure:	read-command-token PORT
; This procedure starts parsing of a command token. The current position
; in the stream must be #\@. This procedure scans enough of the input stream
; to figure out what kind of a command token it is seeing. The procedure returns
; a token structure describing the token. Note, generally reading
; of the current command is not finished! In particular, no arguments of
; the start-tag token are scanned.
;; FIXME: doc return values

(define (read-command-token port)
  (assert-curr-char '(#\@) "start of the command" port)
  (let ((peeked (peek-char port)))
    (cond
     ((memq peeked '(#\! #\. #\? #\@ #\\ #\{ #\}))
      ;; @-commands that escape characters
      (make-token 'STRING (string (read-char port))))
     ((char-alphabetic? peeked)
      (let ((name (read-command port)))
        (cond
         ((eq? name 'end)
          ;; got an ending tag
          (let ((command (string-trim-both
                          (read-eof-line port))))
            (or (and (not (string-null? command))
                     (string-every char-alphabetic? command))
                (parser-error port "malformed @end" command))
            (make-token 'END (string->symbol command))))
         ((eq? name 'bye)
          ;; the end of the top
          (make-token 'END '*TOP*))
         ((eq? name 'item)
          (make-token 'ITEM 'item))
         (else
          (case (cadr (command-spec name))
            ((INLINE-ARGS EOL-ARGS INLINE-TEXT EOL-TEXT ENVIRON
                          EMPTY-COMMAND TABLE-ENVIRON ITEM)
             (make-token 'START name))
            (else
             (parser-error port "can't happen" name))))))))))

; procedure+: 	read-verbatim-body PORT STR-HANDLER SEED
;
; This procedure must be called after we have read a string
; "@verbatim\n" that begins a verbatim section. The current position
; must be the first position of the verbatim body. This function reads
; _lines_ of the verbatim body and passes them to a STR-HANDLER, a
; character data consumer.
;
; The str-handler is a STR-HANDLER, a procedure STRING1 STRING2 SEED.
; The first STRING1 argument to STR-HANDLER never contains a newline.
; The second STRING2 argument often will. On the first invocation of the
; STR-HANDLER, the seed is the one passed to read-verbatim-body
; as the third argument. The result of this first invocation will be
; passed as the seed argument to the second invocation of the line
; consumer, and so on. The result of the last invocation of the
; STR-HANDLER is returned by the read-verbatim-body. Note a
; similarity to the fundamental 'fold' iterator.
;
; Within a verbatim section all characters are taken at their face
; value. It ends with "\n@end verbatim(\r)?\n".

;; Must be called right after the newline after @verbatim.
(define (read-verbatim-body port str-handler seed)
  (let loop ((seed seed))
    (let ((fragment (next-token '() '(#\newline)
                                "reading verbatim" port)))
      ;; We're reading the char after the 'fragment', which is
      ;; #\newline.
      (read-char port)
      (if (string=? fragment "@end verbatim")
          seed
          (loop (str-handler fragment "\n" seed))))))

; The ATTLIST Abstract Data Type
; Currently is implemented as an assoc list sorted in the ascending
; order of NAMES.

(define (make-empty-attlist) '())

(define (command-name-compare sym1 sym2)
  (cond 
   ((eq? symb1 symb2) '=)
   ((string<? (symbol->string symb1) (symbol->string symb2))
    '<)
   (else '>)))

; Add a name-value pair to the existing attlist preserving the order
; Return the new list, in the sorted ascending order.
; Return #f if a pair with the same name already exists in the attlist

(define (attlist-add attlist name-value)
  (if (null? attlist) (cons name-value attlist)
      (case (name-compare (car name-value) (caar attlist))
	((=) #f)
	((<) (cons name-value attlist))
	(else (cons (car attlist) (attlist-add (cdr attlist) name-value)))
	)))

(define attlist-null? null?)

; Given an non-null attlist, return a pair of values: the top and the rest
(define (attlist-remove-top attlist)
  (values (car attlist) (cdr attlist)))

(define (attlist->alist attlist) attlist)
(define attlist-fold fold)

; procedure+:	read-arguments PORT
;
; This procedure reads and parses a production ArgumentList.
; ArgumentList ::= S* Argument (S* , S* Argument)* S*
; Argument ::= ([^@{},])*
;
; Arguments are the things in braces, i.e @ref{my node} has one
; argument, "my node". Most commands taking braces actually don't have
; arguments, they process text. For example, in
; @emph{@strong{emphasized}}, the emph takes text, because the tree
; traversal continues into the braces.
;
; Any whitespace within Argument is replaced with a single space.
; Whitespace around an Argument is trimmed.
;
; The procedure returns a list of arguments. Afterwards the current
; character will be after the final #\}.

(define (read-arguments port stop-char)
  (define delimiters '(#\newline #\return #\space #\tab #\} #\{ #\@ #\,))
  (define (read-argument-helper prev-fragments)
    ;; 'prev-fragments' is the list of string fragments, accumulated so
    ;; far, in reverse order. Return the list of fragments with newly
    ;; read fragments prepended.
    (let* ((new-fragments (cons (next-token '() delimiters
                                            "argument to @-command" port)
                                prev-fragments)))
      (let loop ((cterm (peek-char port)))
        (cond
         ((or (eof-object? cterm)
              (eqv? cterm stop-char)
              (eqv? cterm #\,))
          (if (and (string-null? (car new-fragments))
                   (null? (cdr new-fragments)))
              #f ;; empty argument (i.e. ",,")
              (reverse! new-fragments)))
         ((char-whitespace? cterm)
          (read-char port)
          (let ((peeked (peek-char port)))
            (if (memq peeked delimiters)
                (loop peeked)
                (read-argument-helper (cons " " new-fragments)))))
         (else
          ;; { @
          (parser-error port "Can't have { or @ in a command taking args"))))))
   (define (read-argument)
     (let ((ret (read-argument-helper '())))
       (and=> ret (lambda (x) (apply string-append x)))))

   (skip-horizontal-whitespace port)
   (if (eq? (peek-char port) stop-char)
       (begin (read-char port) '()) ;; No args at all
       (let loop ((arguments (list (read-argument))))
         (let ((c (read-char port)))
           (cond
            ((eq? c #\,)
             (skip-horizontal-whitespace port)
             (loop (cons (read-argument) arguments)))
            ((eq? c stop-char)
             (reverse arguments))
            (else
             (parser-error port "EOF while reading arguments (unclosed brace?)")))))))

; procedure+:	complete-start-command COMMAND PORT
;
; This procedure is to complete parsing of an @-command. The procedure
; must be called after the command token has been read. COMMAND is a
; TAG-NAME.
;
; This procedure returns several values:
;  COMMAND: a symbol.
;  ATTRIBUTES: command's attributes, as an alist.
;  ELEM-CONTENT-MODEL

; On exit, the current position in PORT will depends on the ELEM-CONTENT-MODEL.
; For INLINE-TEXT, it will be one character after the #\{.
; For INLINE-ARGS, it will be the first character after the #\}.
; For EOL-TEXT, it will be the first non-whitespace character after the command.
; For EOL-ARGS, it will be the first character on the next line.
; ENVIRON is like EOL-TEXT.

(define (arguments->attlist port args arg-names)
  (cond
   ((eq? arg-names #t)
    (acons '*arguments* args '()))
   ((not arg-names)
    (or (and (null? args) '())
        (parser-error port "Command doesn't take arguments" arg-names args)))
   (else
    (let loop ((in args) (names arg-names) (out '()))
      (cond
       ((symbol? names) ;; a rest arg
        (reverse (acons names in out)))
       ((null? in)
        (or (and (null? names) (reverse out))
            (parser-error port "@-command expected more arguments:" names)))
       ((null? names)
        (parser-error port "@-command didn't expect more arguments:" in))
       (else
        (loop (cdr in) (cdr names) (acons (car names) (car in) out))))))))

(define (complete-start-command command port)
  (define (get-attributes type arg-names stop-char)
    (arguments->attlist port (read-arguments port stop-char) arg-names))

  (let* ((spec (command-spec command))
         (type (cadr spec))
         (arg-names (cddr spec)))
    (case type
      ((INLINE-TEXT)
       (assert-curr-char '(#\{) "Inline element lacks {" port)
       (values command '() type))
      ((INLINE-ARGS)
       (assert-curr-char '(#\{) "Inline element lacks {" port)
       (values command (get-attributes type arg-names #\}) type))
      ((EOL-TEXT ENVIRON TABLE-ENVIRON)
       (skip-horizontal-whitespace port)
       (values command '() type))
      ((PARAGRAPH EMPTY-COMMAND ITEM)
       (values command '() type))
      ((EOL-ARGS)
       (values command (get-attributes type arg-names #\newline) type))
      (else
       (parser-error port "can't happen")))))

;-----------------------------------------------------------------------------
;			Higher-level parsers and scanners
;
; They parse productions corresponding to the whole (document) entity
; or its higher-level pieces (prolog, root element, etc).


;; Only reads @settitle, leaves it to the command parser to finish
;; reading the title.
(define (take-until-settitle port)
  (or (find-string-from-port? "\n@settitle " port)
      (parser-error port "No \\n@settitle  found"))
  (skip-horizontal-whitespace port)
  (and (eq? (peek-char port) #\newline)
       (parser-error port "You have a @settitle, but no title")))

; procedure+:	read-char-data PORT EXPECT-EOF? STR-HANDLER SEED
;
; This procedure is to read the CharData of a texinfo document.
;
; text ::= (CharData | Command | Environment)*
;
; The procedure reads CharData and stops at @-commands (or
; environments). It also stops at an open or close brace.
;
; port
;	a PORT to read
; expect-eof?
;	a boolean indicating if EOF is normal, i.e., the character
;	data may be terminated by the EOF. EOF is normal
;	while processing the main document.
; str-handler
;	a STR-HANDLER
; seed
;	an argument passed to the first invocation of STR-HANDLER.
;
; The procedure returns two results: SEED and TOKEN.
; The SEED is the result of the last invocation of STR-HANDLER, or the
; original seed if STR-HANDLER was never called.
;
; TOKEN can be either an eof-object (this can happen only if expect-eof?
; was #t), or a texinfo token denoting the start or end of a tag.
;
; @verbatim sections are expanded inline and never returned. Comments
; are silently disregarded.

;; read-char-data port expect-eof? preserve-ws? str-handler seed
(define read-char-data
  (let* ((end-chars-eof '(*eof* #\{ #\} #\@ #\newline)))
    (define (handle-fragment fragment str-handler break? seed)
      (if (and (string-null? fragment) (not break?))
          seed
          (str-handler fragment (if break? "\n" "") seed)))

    (lambda (port expect-eof? preserve-ws? str-handler seed)
      (let ((end-chars ((if expect-eof? identity cdr) end-chars-eof)))
        (let loop ((seed seed))
          (let* ((fragment (next-token '() end-chars "reading char data" port))
                 (term-char (peek-char port))) ; one of end-chars
            (cond
             ((eof-object? term-char) ; only if expect-eof?
              (values (handle-fragment fragment str-handler #f seed)
                      term-char))
             ((memq term-char '(#\@ #\{ #\}))
              (values (handle-fragment fragment str-handler #f seed)
                      (case term-char
                        ((#\@) (read-command-token port))
                        ((#\{) (make-token 'START '*braces*))
                        ((#\}) (read-char port) (make-token 'END #f)))))
             ((eq? term-char #\newline)
              ;; Always significant, unless directly before an end token.
              (let ((c (peek-next-char port)))
                (cond
                 ((eof-object? c)
                  (or expect-eof? (parser-error port "EOF while reading char data"))
                  (values (handle-fragment fragment str-handler #f seed) c))
                 ((eq? c #\@)
                  (let* ((token (read-command-token port))
                         (break? (not (eq? (token-kind token) 'END))))
                    (values (handle-fragment fragment str-handler break? seed)
                            token)))
                 ((and (not preserve-ws?) (eq? c #\newline))
                  ;; paragraph-separator ::= #\newline #\newline+
                  (skip-while '(#\newline) port)
                  (skip-horizontal-whitespace port)
                  (values (handle-fragment fragment str-handler #f seed)
                          (make-token 'PARA 'para)))
                 (else
                  (loop (handle-fragment fragment str-handler #t seed)))))))))))))

; procedure+:	assert-token TOKEN KIND NAME
; Make sure that TOKEN is of anticipated KIND and has anticipated NAME
(define (assert-token token kind name)
  (or (and (token? token)
           (eq? kind (token-kind token))
           (equal? name (token-head token)))
      (parser-error port "Expecting @end for " command
                    ", got " token)))

;========================================================================
;		Highest-level parsers: Texinfo to SXML

; These parsers are a set of syntactic forms to instantiate a SSAX
; parser. The user tells what to do with the parsed character and
; element data. These latter handlers determine if the parsing follows a
; SAX or a DOM model.

; syntax: make-command-parser fdown fup str-handler

; Create a parser to parse and process one element, including its
; character content or children elements. The parser is typically
; applied to the root element of a document.

; fdown
;	procedure COMMAND ATTRIBUTES EXPECTED-CONTENT SEED
;
;	This procedure is to generate the seed to be passed to handlers
;	that process the content of the element. This is the function
;	identified as 'fdown' in the denotational semantics of the XML
;	parser given in the title comments to this file.
;
; fup
;	procedure COMMAND ATTRIBUTES PARENT-SEED SEED
;
;	This procedure is called when parsing of COMMAND is finished.
;	The SEED is the result from the last content parser (or from
;	fdown if the element has the empty content). PARENT-SEED is the
;	same seed as was passed to fdown. The procedure is to generate a
;	seed that will be the result of the element parser. This is the
;	function identified as 'fup' in the denotational semantics of
;	the XML parser given in the title comments to this file.
;
; str-handler
;	A STR-HANDLER
;

; The generated parser is a
;	procedure COMMAND PORT SEED
;
; The procedure must be called *after* the command token has been read.

(define-macro (predicase val . clauses)
  (or (and (pair? clauses) (and-map list? clauses))
      (error "predicase clauses must be lists"))
  (let ((x (gensym)))
    (define (make-valued-p pred)
      (if (eq? pred 'else) pred
          `(if (,pred ,x) ,x #f)))
    `(let ((,x ,val))
       (cond
        ,@(map cons
               (map make-valued-p (map car clauses))
               (map cdr clauses))))))

(define (pmap procs x)
  (map (lambda (p) (p x)) procs))

(define (make-command-parser fdown fup str-handler)
  (lambda (command port seed)
    (define (handle-command command port flag sig-ws? parent-seed)
      (let*-values (((command attributes expected-content)
                     (complete-start-command command port)))
        (define (finish-attr-parsing)
          (cond
           ;; FIXME FIXME: attributes should be handled with a special
           ;; instance of the command parser, that defines its own
           ;; fup/fdown/str-handler, so we can inspect the seed.
           ((eq? expected-content 'ENVIRON)
            (handle-command '*ENVIRON-ATTRS* port #f #t parent-seed))
           ((and (eq? command 'item) (eq? flag 'table))
            (handle-command '*ENVIRON-ATTRS* port #f #t parent-seed))
           ((eq? expected-content 'TABLE-ENVIRON)
            ;; should go in c-s-c
            (let ((line (string-trim-both (read-text-line port))))
              (if (and (not (string-null? line))
                       (eq? (string-ref line 0) #\@))
                  (let* ((how (string->symbol (substring line 1))))
                    (or (inline-content? (cadr (command-spec how)))
                        (parser-error port "Table-how must be INLINE" how))
                    (acons 'how how '()))
                  (if (and (memq command '(enumerate itemize))
                           (case (string-length line)
                             ((0) (eq? command 'enumerate))
                             ((1) (not (char-whitespace? (string-ref line 0))))
                             (else (and (eq? command 'enumerate)
                                        (string-every char-numeric? line)))))
                      (acons 'how line '())
                      (parser-error port "Invalid table-how" line)))))
           (else attributes)))

        (let* ((attributes (finish-attr-parsing))
               (seed (fdown command attributes expected-content parent-seed))
               (eof-closes? (or (memq command '(*TOP* para))
                                (eq? expected-content 'EOL-TEXT)))
               (sig-ws? (or sig-ws? (space-significant? command)))
               (up (lambda (s) (fup command attributes parent-seed s)))
               (down-up (lambda (c port p s) (handle-command c port p sig-ws? s)))
               (new-para (lambda (s) (fdown 'para '() 'PARAGRAPH s)))
               (make-end-para (lambda (p) (lambda (s) (fup 'para '() p s)))))
          
          (define (port-for-content)
            (if (eq? expected-content 'EOL-TEXT)
                (call-with-input-string (read-text-line port) identity)
                port))

          (cond
           ((memq expected-content '(EMPTY-COMMAND INLINE-ARGS EOL-ARGS))
            ;; empty or finished by complete-start-command
            (up seed))
           (else
            (let loop ((port (port-for-content))
                       (expect-eof? eof-closes?)
                       (end-para #f)
                       (need-break? (memq expected-content
                                          '(ENVIRON TABLE-ENVIRON)))
                       (seed seed))
              (cond
               ((and need-break? (or sig-ws? (skip-whitespace port))
                     (not (memq (peek-char port) '(#\@ #\})))
                     (not (eof-object? (peek-char port))))
                ;; Even if we have an @, it might be inline -- check
                ;; that later
                (let ((seed ((or end-para identity) seed)))
                  (loop port expect-eof? (make-end-para seed) #f
                        (new-para seed))))
               (else
                (let*-values (((seed token)
                               (read-char-data
                                port expect-eof? sig-ws? str-handler seed)))
                  (cond
                   ((eof-object? token)
                    (if eof-closes? (up ((or end-para identity) seed)) seed))
                   (else
                    (case (token-kind token)
                      ((STRING)
                       ;; this is only @-commands that escape
                       ;; characters: @}, @@, @{ -- new para if need-break
                       (let ((seed ((or (and need-break? end-para) identity)
                                    seed)))
                         (loop port expect-eof?
                               (and need-break? (make-end-para seed)) #f
                               (str-handler (token-head token) ""
                                            ((if need-break? new-para identity)
                                             seed)))))
                      ((END)
                       ;; The end will only have a name if it's for an
                       ;; environment
                       (cond
                        ((eq? command 'item)
                         (let ((spec (command-spec (token-head token))))
                           (or (eq? (cadr spec) 'TABLE-ENVIRON)
                               (parser-error
                                port "@item not ended by @end table/enumerate/itemize"
                                token))))
                        ((eq? expected-content 'ENVIRON)
                         (assert-token token 'END command)))
                       (up ((or end-para identity) seed)))
                      ((ITEM)
                       (cond
                        ((eq? expected-content 'TABLE-ENVIRON)
                         (up (down-up 'item port command ((or end-para identity) seed))))
                        ((eq? command 'item)
                         (down-up 'item port flag (up ((or end-para identity) seed))))
                        (else
                         (parser-error
                          port "@item must be within a table environment"
                          command))))
                      ((PARA)
                       ;; examine valid paragraphs?
                       (loop port expect-eof? end-para (not sig-ws?) seed))
                      ((INCLUDE)
                       ;; Recurse for include files
                       (let ((seed (call-with-file-and-dir
                                    (token-head token)
                                    (lambda ()
                                      (loop (current-input-port) #t end-para
                                            need-break? seed)))))
                         (loop port expect-eof? end-para need-break? seed)))
                      ((START) ; Start of an @-command
                       (let* ((head (token-head token))
                              (type (cadr (command-spec head)))
                              (inline? (inline-content? type))
                              (seed ((if inline?
                                         (or (and need-break? end-para)
                                             identity)
                                         (or end-para identity)) seed)))
                         (loop port expect-eof? 
                               (and inline?
                                    (if need-break? (make-end-para seed) end-para))
                               (not inline?)
                               (down-up head port #f ((if (and inline? need-break?)
                                                          new-para
                                                          identity) seed)))))
                      (else
                       (parser-error port "Unknown token type" token))))))))))))))

    (handle-command command port #f #f seed)))

; syntax: make-parser user-handler-tag user-handler-proc ...
;
; Create a texinfo parser as an instance of this parsing framework. It
; will be a SAX, a DOM, or a specialized parser depending on the
; handlers that the user supplies.

; user-handler-tag is a symbol that identifies a procedural expression
; that follows the tag. Given below are tags and signatures of the
; corresponding procedures. Not all tags have to be specified. If some
; are omitted, reasonable defaults will apply.

; tag: FDOWN
; handler-procedure: see make-elem-parser, my-new-level-seed

; tag: FUP
; handler-procedure: see make-elem-parser, my-finish-element

; tag: STR-HANDLER
; handler-procedure: see make-elem-parser, my-char-data-handler

; The generated parser is a
;	procedure PORT SEED

(define (make-parser . user-handlers)
  ;; An assoc list of user-handler-tag and default handlers
  (define all-handlers
    `((FDOWN . REQD)                    ; required
      (FUP . REQD)                      ; required
      (STR-HANDLER . REQD)))            ; required

  ;; Delete an association with the tag from alist
  ;; exit to cont, passing the tag, tag-association, and the list
  ;; of remaining associations.
  ;; It's an error if the association with the tag does not exist
  ;; in alist
  (define (delete-assoc alist tag cont)
    (let loop ((alist alist) (scanned '()))
      (cond
       ((null? alist) (error "Unknown user-handler-tag: " tag))
       ((eq? tag (caar alist))
        (cont tag (cdar alist) (append scanned (cdr alist))))
       (else (loop (cdr alist) (cons (car alist) scanned))))))

  ;; create an assoc list of tags and handlers
  ;; based on the defaults and on the given handlers
  (define (merge-handlers declared-handlers given-handlers)
    (cond
     ((null? given-handlers)		; the arguments are exhausted...
      (cond
       ((null? declared-handlers) '())
       ((not (eq? 'REQD (cdar declared-handlers))) ; default value was given:
        (cons (car declared-handlers)	; use it
              (merge-handlers (cdr declared-handlers) given-handlers)))
       (else (error "The handler for the tag " (caar declared-handlers)
                    " must be specified"))))
     ((null? (cdr given-handlers))
      (error "Odd number of arguments to make-parser"))
     (else
      (delete-assoc declared-handlers (car given-handlers)
                    (lambda (tag value alist)
                      (cons (cons tag (cadr given-handlers))
                            (merge-handlers alist (cddr given-handlers))))))))

  (let ((user-handlers (merge-handlers all-handlers user-handlers)))

    (define (get-handler tag)
      (cond
       ((assq tag user-handlers) => cdr)
       (else (error "unknown tag: " tag))))

    ;; this is the parser
    (lambda (port seed)
      
      ;; A procedure command port seed
      (define command-parser
        (make-command-parser (get-handler 'FDOWN)
                             (get-handler 'FUP)
                             (get-handler 'STR-HANDLER)))

      (take-until-settitle port)
      (command-parser '*TOP* port seed))))


;========================================================================
;		Highest-level parsers: Texinfo to SXML
;

; procedure: texinfo->sxml PORT NAMESPACE-PREFIX-ASSIG
;
; This is an instance of a SSAX parser above that returns an SXML
; representation of the texinfo document ready to be read at PORT.
;
; The procedure returns an SXML tree. The port points to the
; first character after the @bye, or to the end of the file.

(define (texinfo->sxml port)
  (define (reverse-collect-str-drop-ws fragments)
    ;; given the list of fragments (some of which are text strings)
    ;; reverse the list and concatenate adjacent text strings We also
    ;; drop "unsignificant" whitespace, that is, whitespace in front,
    ;; behind and between elements. The whitespace that is included in
    ;; character data is not affected.
    (cond 
     ((null? fragments) '())		; a shortcut
     ((and (string? (car fragments))	; another shortcut
           (null? (cdr fragments))	; remove trailing ws
           (string-whitespace? (car fragments))) '())
     (else
      (let loop ((fragments fragments) (result '()) (strs '())
                 (all-whitespace? #t))
        (cond
         ((null? fragments)
          (if all-whitespace? result	; remove leading ws
              (cons (apply string-append strs) result)))
         ((string? (car fragments))
          (loop (cdr fragments) result (cons (car fragments) strs)
                (and all-whitespace?
                     (string-whitespace? (car fragments)))))
         (else
          (loop (cdr fragments)
                (cons
                 (car fragments)
                 (if all-whitespace? result
                     (cons (apply string-append strs) result)))
                '() #t)))))))

    (let ((result
	   (reverse
	    ((make-parser
              'FDOWN
              (lambda (command attributes expected-content seed)
                ;;(pk 'down command)
                ;;(pk 'fdown command attributes expected-content seed)
                '())
   
              'FUP
              (lambda (command attributes parent-seed seed)
                ;;(pk 'up command)
                ;;(pk 'fup command attributes parent-seed seed)
                (let ((seed (reverse-collect-str-drop-ws seed))
                      (attrs
                       (attlist-fold
                        (lambda (attr accum)
                          (cons (list (car attr) (cdr attr))
                                accum))
                        '()
                        (if (and (not (null? attributes))
                                 (eq? (caar attributes) '*ENVIRON-ATTRS*))
                            (arguments->attlist
                             port (cdar attributes)
                             (cddr (assq command texinfo-command-specs)))
                            attributes))))
                    (cons
                     (cons 
                      command
                      (if (null? attrs)
                          seed
                          (cons (cons '@ attrs) seed)))
                     parent-seed)))

              'STR-HANDLER
              (lambda (string1 string2 seed)
                ;;(pk 'str-handler string1 string2 seed)
                (if (string-null? string2)
                    (cons string1 seed)
                    (cons* string2 string1 seed))))
             port '()))))
      (postprocess (car result))))

(define (postprocess tree)
  (define (trim-whitespace str trim-left? trim-right?)
    (let loop ((in (string->list str)) (out '()) (spaced? #f))
      (cond
       ((null? in)
        (let* ((out (if trim-right? (drop-while char-whitespace? out) out))
               (reversed (reverse out))
               (trimmed (if trim-left? (drop-while char-whitespace? reversed) reversed)))
          (list->string trimmed)))
       ((char-whitespace? (car in))
        (loop (cdr in) (if spaced? out (cons #\space out)) #t))
       (else
        (loop (cdr in) (cons (car in) out) #f)))))
  (define (remove-whitespace tree)
    (cond
     ((null? tree)
      tree)
     ((string? tree)
      (trim-whitespace tree))
     ((space-significant? (car tree))
      tree)
     (else
      (let lp ((kids (cdr tree)) (out '()) (first? #t))
        (cond
         ((null? kids) (cons (car tree) (reverse out)))
         ((pair? (car kids))
          (if (eq? (caar kids) '@)
              (lp (cdr kids) (cons (car kids) out) #t)
              (lp (cdr kids) (cons (remove-whitespace (car kids)) out) #f)))
         (else ;; string
          (lp (cdr kids) 
              (cons (trim-whitespace (car kids) first? (null? (cdr kids)))
                    out)
              #f)))))))
  (remove-whitespace tree)
  )
;;; arch-tag: 73890afa-597c-4264-ae70-46fe7756ffb5
;;; texinfo.scm ends here
